#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""


Other packages that might be interesting to the reader:

    billard
    lightqueue
    pandaemonium
    papy
    parallel_run
    pistil
    PY4GRID
    pyjobs
    PyroMP
    pyumpf
    swarm
    zprocess
    alchemy
    async_gui
    distprocessing
    FsQueue
    lazypy
    lelo
    mycloud
    parallel-queue
    parallelize
    smartpool
"""
from __future__ import division, print_function

import os
import time
import multiprocessing as mp
import uilayer


__all__= ["ProcessPool"]


class ProcessPool(object):
    def __init__(self, targets, args, user_interface=None, kwargs=None, 
                 maxNumOfRunningJobs=0, lock=None, quietPool=True,
                 updateInterval_sec=.1):
        """ Multiprocessing Progress Manager
        
        Run a couple of functions in parallel. Optionally allows
        to output the progress of the functions. For this, the
        functions need to 
        
        
        Parameters
        ----------
        targets : list of callable
            The functions that are to be evaluated
        args : list of tuples of arguments
            The arguments to each function
        user_interface : instance of `ttui.base_ui`, optional
            User interface for progress report
        kwargs : dict, optional
            Keyword arguments for the user interface methods
            possible arguments:
                kwargs["name"] : string to display when finished
                kwargs["steps"] : integer used to compute percentage
            Do not use these arguments:
                kwargs["pid"] : integer used to identify process
                kwargs["trigger"] : trigger function for `targets`
        maxNumOfRunningJobs : int, optional
            Defines the number of processes that are run at a time.
            Negative values are interpreted as number of unnused
            cores. The number of available cores is automatically
            determined.
        lock : instance of `mp.Lock`, optional
        quietPool : boolean, optional
            If set to True, the ProcessPool will be verbose.
        updateInterval  : float, optional
            Set the time interval with which we check for finished
            jobs and which we use to update the user interface.
    
    
        Methods
        -------
        run : callable
            Starts executing the `targets`.
            

        TODO
        ----
        Scan function arguments for pointers and create shared
        memory for processes.
        if args[0][1] is args[1][1]:
            create_shared_memory_for args[i][1]
        
        `targets` or `args` can just be tuples. In that case we
        could make the lists ourselves or create shared objects...
        or manager.Namespace
            
        """
        self._targetList = targets
        self._argsList = args
        if kwargs is None:
            kwargs = [{}] * len(args)
        self._kwargsList = kwargs
        
        if maxNumOfRunningJobs <= 0:
            nCPU = mp.cpu_count()
            self._maxNumOfRunningJobs = nCPU + maxNumOfRunningJobs
            
            if not quietPool:
                print("...maxNumOfRunningJobs is <= 0 (maxRun {})".
                      format(maxNumOfRunningJobs))
                print("...I interpret that as number of unused cores.")
                print("....numCores {}".format(nCPU))
            #have at least one job running
            if self._maxNumOfRunningJobs < 1:
                print("....resulting cores to use is non positive ({})".
                      format(self._maxNumOfRunningJobs))
                print("....so maxRun will be set to 1")
                self._maxNumOfRunningJobs = 1
                
        else:
            self._maxNumOfRunningJobs = maxNumOfRunningJobs
        
        if not quietPool:
            print("....maxRun {}".format(self._maxNumOfRunningJobs))
            
        self._quiet = quietPool
        self._totalNumberOfJobs = len(self._argsList)
        self._jobsDone = mp.Value('i', 0)
        self._jobsRunning = mp.Value('i', 0)
        self._jobList = []
        self._lock = lock
        self._manager = mp.Manager()
        self._results = self._manager.dict()
        self._map_pid2name = dict()
        
        self._pids = mp.Array('i', [-1]*self._maxNumOfRunningJobs)
        self._progress = mp.Array('f', self._maxNumOfRunningJobs)
        self._steps = mp.Array('f', self._maxNumOfRunningJobs)

        ## Set the user interface
        self.ui = user_interface
        if self.ui is not None:
            # give user interface data
            self.ui.SetProgress(self._steps, self._pids, self._progress)
        
        self._updateInterval_sec = updateInterval_sec
    
        for i in range(len(self._argsList)):
            args = self._argsList[i]
            kwargs = self._kwargsList[i]
            self._jobList.append(mp.Process(
                                     target = self.__targetWrapper,
                                     args = (self._jobsDone,
                                             self._jobsRunning,
                                             self._targetList[i])+ args,
                                     kwargs = kwargs,
                                     name = i+1
                                         ),
                                  )


    def __targetWrapper(self, _jobsDone, _jobsRunning, target,
                        *args, **kwargs):
        pid = os.getpid()
        if not self._quiet:
            print("...starting job PID {}".format( pid ))
            
        for i in range(len(self._pids)):
            if self._pids[i] == -1:
                self._pids[i] = pid
                nstep = kwargs.get("steps")
                if nstep is None:
                    nstep = 0
                self._steps[i] = nstep
                break

        if self.ui is not None:
            func = self.ui.Iterate
        else:
            func = None
        
        newargs = args
        
        # Add an additional keyword argument for the trigger function
        kwargs["pid"] = pid
        result = target(*newargs, trigger=func, trigger_kwargs=kwargs)
        self._results[pid] = result

        for i in range(len(self._pids)):
            if self._pids[i] == pid:
                self._pids[i] = -1
                self._progress[i] = 0

        if self.ui is not None:
            self.ui.Finalize(**kwargs)

        _jobsDone.value += 1
        _jobsRunning.value -= 1

        if not self._quiet:
            print("...finished job #{}".format(os.getpid()))


    def __triggerNewJobs(self):
        while ((self._jobsRunning.value < self._maxNumOfRunningJobs)
               and (len(self._jobList) > 0)):
            job = self._jobList.pop()
            if not self._quiet:
                print("...triggered job #{}".format(
                          self._totalNumberOfJobs - len(self._jobList)))
            job.start()
            self._map_pid2name[job.pid] = job.name
            self._jobsRunning.value += 1


    def get_results(self):
        """
            Return the results in a list corresponding to the input
            data.
        """
        results = list()
        names = list()
        for pid in list(self._results.keys()):
            results.append(self._results[pid])
            names.append(self._map_pid2name[pid])

        # sort results according to names
        results = [x for (y,x) in sorted(zip(names,results))]
        
        return results



    def run(self):
        """
            Execute the `targets` that have been defined during the
            init process.
        """
        stop = False
        while not stop:
            self.__triggerNewJobs()

            time.sleep(self._updateInterval_sec)
            stop = (self._jobsDone.value == self._totalNumberOfJobs)

            if self.ui is not None:
                self.ui.Update()


def _test():
    def prcs(arg, N=50, trigger=None, trigger_kwargs={}):
        """
            Counts from 1 to N and sleeps in between for arg*0.01 s.
            Calls the function `trigger` with **kwargs after each
            wakeup.
        """
        a = time.time()
        t = .01*arg
        for i in range(N):
            time.sleep(t)
            if trigger is not None:
                trigger(**trigger_kwargs)
        return time.time() - a
    
    N = 100
    
    NumCalls = 11
    
    targets = [prcs] * NumCalls
    targ_args = [(i, N) for i in range(1,NumCalls+1)]

    ui = uilayer.stdout()
    ui_kwargs = [{"name" : "waiting game", "steps" : N}] * NumCalls
    pool = ProcessPool(targets=targets, args=targ_args,
                       user_interface=ui, kwargs=ui_kwargs,
                       maxNumOfRunningJobs=0, lock=mp.Lock(),
                       quietPool=True)
    pool.run()

    results = pool.get_results()
    for i in range(len(results)):
        print("Result of target #{}: {}".format(i+1, results[i]))


__version__ = "0.1.0"
__author__ = "Paul Mueller"
__email__ = "paul.mueller@biotec.tu-dresden.de"
__license__ = "BSD (3-Clause)"

